/*
 * Diskheap IO defines...
 *
 * Copyright 2002 by Gray Watson
 *
 * This file is part of the diskheap package.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Gray Watson not be used in advertising
 * or publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be contacted via http://256.com/gray/
 *
 * $Id: diskheap_io.h,v 1.14 2002/02/19 15:32:33 gray Exp $
 */

#ifndef __DISKHEAP_IO_H__
#define __DISKHEAP_IO_H__

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

/*
 * This is the default version number of the i/o routines.  The idea
 * is that whenever I change the on-disk formats, I'll rotate the
 * version number and write support functions for the old versions.
 */
#define DISKHEAP_DEFAULT_VERSION	1

#define DISKHEAP_FILE_MAGIC	"GWDH2002"
#define DISKHEAP_FILE_MAGIC_LEN	(sizeof(DISKHEAP_FILE_MAGIC) - 1)

/*
 * Maximum header size.  We set this not to the block-size because the
 * user may have set some huge block-size for the heap and we don't
 * want to waste that much.  The free-list information will be stored
 * immediately past the header information.
 */
#define DISKHEAP_MAX_HEADER	512

/*
 * This is how much free stuff to process at any one time.  Since the
 * user might select some large block-size, we don't use that
 * variable.
 */
#define DISKHEAP_FREE_CHUNK	512

/*
 * Default start of the user allocation sections.  We use the first
 * 1mb to store administrative and free-space information.  This can
 * be extended, immediately after the diskheap is created with the
 * diskheap_extend_freespace command.
 */
#define DISKHEAP_DEFAULT_USER_START	(1024 * 1024)

/*
 * Block types
 */
#define BLOCK_TYPE_USER_ALL 	1	/* contiguous user allocation */
#define BLOCK_TYPE_FREE 	2	/* free space */
#define BLOCK_TYPE_SUB	 	3	/* sub-block allocation */

#define BLOCK_MAGIC_NUMBER	0x2BED	/* hex magic block header */

/*
 * structure which holds all of the various information from a block
 * header
 */
typedef struct {
  unsigned int	bh_magic;		/* here as a placeholder */
  unsigned int	bh_block_type;		/* type of the block - see above */
  unsigned int	bh_user_type;		/* type given by the user */
  unsigned int	bh_block_n;		/* number of blocks used on disk */
  unsigned int	bh_user_size;		/* real size of user alloc in bytes */
} block_head_t;

/*
 * This seems like it would be pretty portable.  Basically we are
 * trying to find out how many bytes would be in the maximum block or
 * sub-block header size.  Since all of the fields in the block_head_t
 * structure are encoded and put into the header, we take the size of
 * it times 8 to get the number of bits, add in 7-1 to handle integer
 * rounding, and divide by 7 to get the number of bits it will encoded
 * into.  This will be adequate.
 */
#define BLOCK_MAX_HEAD_LEN	((sizeof(block_head_t) * 8 + (7 - 1)) / 7)

/*<<<<<<<<<<  The below prototypes are auto-generated by fillproto */

/*
 * int _dh_read_header
 *
 * DESCRIPTION:
 *
 * Read the disk header from a diskheap file.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
extern
int	_dh_read_header(diskheap_t *dh_p);

/*
 * int _dh_write_header
 *
 * DESCRIPTION:
 *
 * Write the header into a disk heap.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
extern
int	_dh_write_header(diskheap_t *dh_p);

/*
 * int _dh_read_free
 *
 * DESCRIPTION:
 *
 * Read the free list from a diskheap file.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
extern
int	_dh_read_free(diskheap_t *dh_p);

/*
 * int _dh_write_free
 *
 * DESCRIPTION:
 *
 * Write out the free list to the diskheap file.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
extern
int	_dh_write_free(diskheap_t *dh_p);

/*
 * int _dh_write_initial_user
 *
 * DESCRIPTION:
 *
 * Write a block of 0s at the start of the user allocation section.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 */
extern
int	_dh_write_initial_user(diskheap_t *dh_p);

/*
 * int _dh_block_header
 *
 * DESCRIPTION:
 *
 * Write the block header information into the start of the internal
 * block buffer.
 *
 * RETURNS:
 *
 * Success - Number of bytes written into the header buffer.
 *
 * Failure - -1
 *
 * ARGUMENTS:
 *
 * buf <-> Character buffer into which we will be writing the header.
 *
 * bounds_p <-> Pointer past the end of the character buffer.
 *
 * head_p -> Pointer to the header information we are storing in the
 * block.
 */
extern
int	_dh_block_header(unsigned char *buf, const unsigned char *bounds_p,
			 const block_head_t *head_p);

/*
 * int _dh_sub_block_header
 *
 * DESCRIPTION:
 *
 * Write the sub-block header information into the start of the
 * internal block buffer.
 *
 * RETURNS:
 *
 * Success - Number of bytes written into the header buffer.
 *
 * Failure - -1
 *
 * ARGUMENTS:
 *
 * buf <-> Character buffer into which we will be writing the header.
 *
 * bounds_p <-> Pointer past the end of the character buffer.
 *
 * head_p -> Pointer to the header information we are storing in the
 * sub-block.
 */
extern
int	_dh_sub_block_header(unsigned char *buf, const unsigned char *bounds_p,
			     const block_head_t *head_p);

/*
 * int _dh_write_header_user
 *
 * DESCRIPTION:
 *
 * Write the header in the dh buffer and the user area to block(s).
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 *
 * head_len -> Length in the header in bytes in the dh_p block buffer.
 *
 * buffer -> User's anonymous memory to be written to disk.
 *
 * user_size -> Size of the user memory.  Set to 0 to just write out a
 * new header.
 *
 * block_num -> Block-number were we are writing it.  No offset
 * necessary since we are writing a block header here.
 */
extern
int	_dh_write_header_user(diskheap_t *dh_p, const int head_len,
			      const unsigned char *buffer,
			      const unsigned int user_size,
			      const unsigned int block_num);

/*
 * int _dh_read_first_block
 *
 * DESCRIPTION:
 *
 * Read in a block header plus a certain number of bytes of the user
 * space up to a block-size worth into the internal buffer space.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 *
 * block_num -> Block-number were we are reading from.
 *
 * max_read_size -> Set the maximum number of bytes to read from the
 * data item or 0 to read all of the bytes.
 *
 * head_p <-> Pointer to a header structure which will be set with the
 * details about the block.
 *
 * head_len_p <- Pointer to an integer which, if not NULL, will be set
 * to the length of the header at the front of the block.
 */
extern
int	_dh_read_first_block(diskheap_t *dh_p, const unsigned int block_num,
			     const unsigned int max_read_size,
			     block_head_t *head_p, int *head_len_p);

/*
 * int _dh_read_sub_block
 *
 * DESCRIPTION:
 *
 * Read in a sub-block.  We first read in and verify the front of the
 * block then we verify and move down to the offset in question.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are working on.
 *
 * block_num -> Block-number were we are reading from.
 *
 * offset -> Offset where we are writing.
 *
 * buffer -> Buffer to store the data in or NULL to alloc.
 *
 * max_read_size -> Set the maximum number of bytes to read from the
 * data item or 0 to read all of the bytes.
 *
 * head_p <-> Pointer to a header structure which will be set with the
 * details about the block.
 *
 * slot_size_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set to the number of bytes in the sub-block slot-size.
 *
 * user_size_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set to the number of bytes in the user's on-disk chunk, not
 * including any internal header information.
 *
 * user_type_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set to the type value which was passed to the store
 * function.
 *
 * error_p <- Pointer to an integer which, if not NULL, will be set
 * with a diskheap error code.
 */
extern
void	*_dh_read_sub_block(diskheap_t *dh_p, const unsigned int block_num,
			    const unsigned int offset,
			    void *buffer, const unsigned int max_read_size,
			    unsigned int *slot_size_p,
			    unsigned int *user_size_p,
			    unsigned int *user_type_p, int *error_p);

/*<<<<<<<<<<   This is end of the auto-generated output from fillproto. */

#ifdef __cplusplus
}
#endif


#endif /* ! __DISKHEAP_IO_H__ */
