/*
 * Diskheap Library Operating System Specific Routines...
 *
 * Copyright 2002 by Gray Watson
 *
 * This file is part of the diskheap package.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Gray Watson not be used in advertising
 * or publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be contacted via http://256.com/gray/
 *
 * $Id: diskheap_os.c,v 1.9 2002/02/19 15:32:34 gray Exp $
 */

#if HAVE_FCNTL_H
#include <fcntl.h>
#endif
#if HAVE_TIME_H
#include <time.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_TIME_H
#include <time.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_SYS_TYPES_H
#include <sys/types.h>			/* for off_t, etc. */
#endif

#ifdef DMALLOC
#include "dmalloc.h"
#endif

#define __DISKHEAP_MAIN__

#include "conf.h"
#include "diskheap.h"
#include "diskheap_os.h"

#if IDENT_WORKS
#ident "$Id: diskheap_os.c,v 1.9 2002/02/19 15:32:34 gray Exp $"
#else
static char *rcs_id = "$Id: diskheap_os.c,v 1.9 2002/02/19 15:32:34 gray Exp $";
#endif

#if SIZEOF_OFF_T == 8

/* YES!! */
#undef OS_OFFSET_32_BIT
typedef off_t	_os_off_t;

#else /* SIZEOF_OFF_T != 8 */
# if OFF64_T_EXISTS

/* well, if that's the best that they can do... */
#undef OS_OFFSET_32_BIT
typedef off64_t	_os_off_t;

# else /* ! OFF64_T_EXISTS */

/* we are using a 32-bit off_t -- disappointed */
#define OS_OFFSET_32_BIT 1
typedef off_t	_os_off_t;

# endif /* ! OFF64_T_EXISTS */
#endif /* SIZEOF_OFF_T != 8 */

/*
 * int _dh_os_open
 *
 * DESCRIPTION:
 *
 * Perform an open call.  Should not be necessary except for crappy
 * implementation of 64-bit filesystems.
 *
 * RETURNS:
 *
 * Success - File description >= 0.
 *
 * Failure - < 0
 *
 * ARGUMENTS:
 *
 * file -> Name of the file we are opening.
 *
 * create_b -> 1 if we are creating the file otherwise 0 if we are
 * just opening it.
 *
 * read_only_b -> 1 if we are opening in read-only mode.  Ignored if
 * creating.
 *
 * mode -> The mode to create the file if we are creating.
 */
int	_dh_os_open(const char *file, const int create_b,
		    const int read_only_b, const int mode)
{
  int	fd, flags;
  
  /* the file can't be there already */
  if (create_b) {
    flags = O_RDWR | O_CREAT | O_EXCL;
  }
  else if (read_only_b) {
    flags = O_RDONLY;
  }
  else {
    flags = O_RDWR;
  }
  
  fd = open(file, flags, mode);
  
  return fd;
}

/*
 * void _dh_os_close
 *
 * DESCRIPTION:
 *
 * Close a file descriptor.
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * fd -> File descriptor to close.
 */
void	_dh_os_close(const int fd)
{
  (void)close(fd);
}

/*
 * int _dh_os_pread
 *
 * DESCRIPTION:
 *
 * Perform a pread using 64-bit (if available) information.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are reading from.
 *
 * buf -> Buffer into which we are reading.
 *
 * bytes_n -> Number of bytes to be read.
 *
 * block_num -> Number of the block where we are reading.
 *
 * offset -> Offset in the block where we are reading.  This can also
 * be used as a 32-bit offset if the block-number if 0.
 */
int	_dh_os_pread(diskheap_t *dh_p, void *buf, const unsigned int bytes_n,
		     const unsigned int block_num, const unsigned int offset)
{
  _os_off_t	position = 0;
  int		ret;
  
  if (block_num > 0) {
    /* need to determine if we have a 32-bit off_t or a 31-bit signed */
    if ((_os_off_t)block_num > OFF_T_MAX / (_os_off_t)dh_p->dh_block_size) {
      return DISKHEAP_ERROR_OFFSET_BOUNDS;
    }
    position = (_os_off_t)block_num * (_os_off_t)dh_p->dh_block_size;
  }
  position += offset;
  
#if HAVE_PREAD64
  ret = pread64(dh_p->dh_fd, buf, bytes_n, position);
#else
# if HAVE_PREAD
  ret = pread(dh_p->dh_fd, buf, bytes_n, position);
# else
  {
    _os_off_t	seek_ret;
    
    seek_ret = lseek(dh_p->dh_fd, position, SEEK_SET);
    if (seek_ret != position) {
      return DISKHEAP_ERROR_LSEEK;
    }
  }
  ret = read(dh_p->dh_fd, buf, bytes_n);
# endif
#endif
  
  if (ret < 0) {
    return DISKHEAP_ERROR_READ;
  }
  else if ((unsigned int)ret == bytes_n) {
    return DISKHEAP_ERROR_NONE;
  }
  else {
    return DISKHEAP_ERROR_READ;
  }
}

/*
 * int _dh_os_pwrite
 *
 * DESCRIPTION:
 *
 * Perform a pwrite using 64-bit (if available) information.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Disk pointer we are reading from.
 *
 * buf -> Buffer from which we are writing.
 *
 * bytes_n -> Number of bytes to be writing.
 *
 * block_num -> Number of the block where we are writing.
 *
 * offset -> Offset in the block where we are writing.
 */
int	_dh_os_pwrite(diskheap_t *dh_p, const void *buf,
		      const unsigned int bytes_n, const unsigned int block_num,
		      const unsigned int offset)
{
  _os_off_t	position = 0;
  int		ret;
  
  if (block_num > 0) {
    /* need to determine if we have a 32-bit off_t or a 31-bit signed */
    if ((_os_off_t)block_num > OFF_T_MAX / (_os_off_t)dh_p->dh_block_size) {
      return DISKHEAP_ERROR_OFFSET_BOUNDS;
    }
    position = (_os_off_t)block_num * (_os_off_t)dh_p->dh_block_size;
  }
  position += offset;
  
#if HAVE_PWRITE64
  ret = pwrite64(dh_p->dh_fd, buf, bytes_n, position);
#else
# if HAVE_PWRITE
  ret = pwrite(dh_p->dh_fd, buf, bytes_n, position);
# else
  {
    _os_off_t	seek_ret;
    
    seek_ret = lseek(dh_p->dh_fd, position, SEEK_SET);
    if (seek_ret != position) {
      return DISKHEAP_ERROR_LSEEK;
    }
  }
  ret = write(dh_p->dh_fd, buf, bytes_n);
# endif
#endif
  
  if (ret < 0) {
    return DISKHEAP_ERROR_WRITE;
  }
  else if ((unsigned int)ret == bytes_n) {
    return DISKHEAP_ERROR_NONE;
  }
  else {
    return DISKHEAP_ERROR_WRITE;
  }
}

/*
 * int _dh_os_srandom
 *
 * DESCRIPTION:
 *
 * Seed the random number generator.
 *
 * RETURNS:
 *
 * The seed chosen by the routine.
 *
 * ARGUMENTS:
 *
 * seed -> Seed specified by the user.  0 for none.
 */
int	_dh_os_srandom(const int user_seed)
{
  int	seed;
  
  if (user_seed != 0) {
    seed = user_seed;
  }
  else {
    seed = time(NULL);
#if HAVE_GETPID
    seed ^= getpid();
#endif
  }
  
#if HAVE_SRANDOM
  srandom(seed);
#else
#if HAVE_SRAND
  srand(seed);
#else
  must have either srandom() or srand() functions;
#endif
#endif
  
  return seed;
}

/*
 * int _dh_os_random
 *
 * DESCRIPTION:
 *
 * Get a random number.
 *
 * RETURNS:
 *
 * A pseudo-random number between 0 and 2^31-1.
 *
 * ARGUMENTS:
 *
 * None.
 */
int	_dh_os_random(void)
{
  int	ret = 0;
  
#if HAVE_RANDOM
  ret = random();
#else
#if HAVE_SRAND
  ret = rand();
#else
  must have either random() or rand() functions;
#endif
#endif
  
  return ret;
}

/*
 * unsigned int _dh_os_time
 *
 * DESCRIPTION:
 *
 * Get the time in seconds.
 *
 * RETURNS:
 *
 * The time.
 *
 * ARGUMENTS:
 *
 * None.
 */
unsigned int	_dh_os_time(void)
{
  unsigned int	ret = 0;
  
#if HAVE_TIME
  ret = time(NULL);
#endif
  
  return ret;
}

/*
 * int _dh_os_fsync
 *
 * DESCRIPTION:
 *
 * Sync the file descriptor if fsync is enabled by the system.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * fd -> File descriptor we are syncing.
 */
int	_dh_os_fsync(const int fd)
{
#if HAVE_FSYNC
  /* call fsync */
  if (fsync(fd) == 0) {
    return DISKHEAP_ERROR_NONE;
  }
  else {
    return DISKHEAP_ERROR_FSYNC;
  }
#else
  return DISKHEAP_ERROR_OS_CAPABLE;
#endif
}
