/*
 * Diskheap Library Test Program...
 *
 * Copyright 2002 by Gray Watson
 *
 * This file is part of the diskheap package.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Gray Watson not be used in advertising
 * or publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be contacted via http://256.com/gray/
 *
 * $Id: diskheap_t.c,v 1.22 2002/02/19 16:13:51 gray Exp $
 */

#include <errno.h>
#include <stdio.h>
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_TIME_H
#include <time.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef DMALLOC
#include "dmalloc.h"
#endif

#include "conf.h"
#include "diskheap.h"
#include "diskheap_os.h"		/* for [s]random functions */

#if IDENT_WORKS
#ident "$Id: diskheap_t.c,v 1.22 2002/02/19 16:13:51 gray Exp $"
#else
static char *rcs_id = "$Id: diskheap_t.c,v 1.22 2002/02/19 16:13:51 gray Exp $";
#endif

#define TEST_DB_FILE		"test.dh"	/* test diskheap file name */
#define DEFAULT_ITERATIONS	1000		/* number of tests to run */
#define MAX_ALLOCATION		(1024*1024)	/* max allocation size */

typedef struct test_slot_st {
  unsigned int		ts_block_num;		/* block-number where stored */
  unsigned int		ts_offset;		/* offset where it is stored */
  unsigned int		ts_user_type;		/* type we gave it */
  unsigned int		ts_data_size;		/* size of random data */
  unsigned int		ts_data_hash;		/* hash of our data */
  struct test_slot_st	*ts_next_p;		/* linked list */
} test_slot_t;

/* local variables */
static	char	*program;		/* argv[0] value */
static	int	silent_b = 0;		/* silent (no) messages */
static	int	verbose_b = 0;		/* verbose messages */
static	FILE	*error_stream;		/* destination of error messages */

/*
 * void HASH_MIX
 *
 * DESCRIPTION:
 *
 * Mix 3 32-bit values reversibly.  For every delta with one or two
 * bits set, and the deltas of all three high bits or all three low
 * bits, whether the original value of a,b,c is almost all zero or is
 * uniformly distributed.
 *
 * If HASH_MIX() is run forward or backward, at least 32 bits in a,b,c
 * have at least 1/4 probability of changing.  If mix() is run
 * forward, every bit of c will change between 1/3 and 2/3 of the
 * time.  (Well, 22/100 and 78/100 for some 2-bit deltas.)
 *
 * HASH_MIX() takes 36 machine instructions, but only 18 cycles on a
 * superscalar machine (like a Pentium or a Sparc).  No faster mixer
 * seems to work, that's the result of my brute-force search.  There
 * were about 2^68 hashes to choose from.  I only tested about a
 * billion of those.
 */
#define HASH_MIX(a, b, c) \
 do { \
   a -= b; a -= c; a ^= (c >> 13); \
   b -= c; b -= a; b ^= (a << 8); \
   c -= a; c -= b; c ^= (b >> 13); \
   a -= b; a -= c; a ^= (c >> 12); \
   b -= c; b -= a; b ^= (a << 16); \
   c -= a; c -= b; c ^= (b >> 5); \
   a -= b; a -= c; a ^= (c >> 3); \
   b -= c; b -= a; b ^= (a << 10); \
   c -= a; c -= b; c ^= (b >> 15); \
 } while(0)

/*
 * static unsigned int hash
 *
 * DESCRIPTION:
 *
 * Hash a variable-length key into a 32-bit value.  Every bit of the
 * key affects every bit of the return value.  Every 1-bit and 2-bit
 * delta achieves avalanche.  About (6 * len + 35) instructions.  The
 * best hash table sizes are powers of 2.  There is no need to use mod
 * (sooo slow!).  If you need less than 32 bits, use a bitmask.  For
 * example, if you need only 10 bits, do h = (h & hashmask(10)); In
 * which case, the hash table should have hashsize(10) elements.
 *
 * By Bob Jenkins, 1996.  bob_jenkins@compuserve.com.  You may use
 * this code any way you wish, private, educational, or commercial.
 * It's free.  See
 * http://ourworld.compuserve.com/homepages/bob_jenkins/evahash.htm
 * Use for hash table lookup, or anything where one collision in 2^^32
 * is acceptable.  Do NOT use for cryptographic purposes.
 *
 * RETURNS:
 *
 * Returns a 32-bit hash value.
 *
 * ARGUMENTS:
 *
 * key - Key (the unaligned variable-length array of bytes) that we
 * are hashing.
 *
 * length - Length of the key in bytes.
 *
 * init_val - Initialization value of the hash if you need to hash a
 * number of strings together.  For instance, if you are hashing N
 * strings (unsigned char **)keys, do it like this:
 *
 * for (i=0, h=0; i<N; ++i) h = hash( keys[i], len[i], h);
 */
static	unsigned int	hash(const unsigned char *key,
			     const unsigned int length,
			     const unsigned int init_val)
{
  const unsigned char	*key_p = key;
  unsigned int		a, b, c, len;
  
  /* set up the internal state */
  a = 0x9e3779b9;	/* the golden ratio; an arbitrary value */
  b = 0x9e3779b9;
  c = init_val;		/* the previous hash value */
  
  /* handle most of the key */
  for (len = length; len >= 12; len -= 12) {
    a += (key_p[0]
	  + ((unsigned int)key_p[1] << 8)
	  + ((unsigned int)key_p[2] << 16)
	  + ((unsigned int)key_p[3] << 24));
    b += (key_p[4]
	  + ((unsigned int)key_p[5] << 8)
	  + ((unsigned int)key_p[6] << 16)
	  + ((unsigned int)key_p[7] << 24));
    c += (key_p[8]
	  + ((unsigned int)key_p[9] << 8)
	  + ((unsigned int)key_p[10] << 16)
	  + ((unsigned int)key_p[11] << 24));
    HASH_MIX(a,b,c);
    key_p += 12;
  }
  
  c += length;
  
  /* all the case statements fall through to the next */
  switch(len) {
  case 11:
    c += ((unsigned int)key_p[10] << 24);
  case 10:
    c += ((unsigned int)key_p[9] << 16);
  case 9:
    c += ((unsigned int)key_p[8] << 8);
    /* the first byte of c is reserved for the length */
  case 8:
    b += ((unsigned int)key_p[7] << 24);
  case 7:
    b += ((unsigned int)key_p[6] << 16);
  case 6:
    b += ((unsigned int)key_p[5] << 8);
  case 5:
    b += key_p[4];
  case 4:
    a += ((unsigned int)key_p[3] << 24);
  case 3:
    a += ((unsigned int)key_p[2] << 16);
  case 2:
    a += ((unsigned int)key_p[1] << 8);
  case 1:
    a += key_p[0];
    /* case 0: nothing left to add */
  }
  HASH_MIX(a, b, c);
  
  return c;
}

/*
 * static void create_block
 *
 * DESCRIPTION:
 *
 * Create a random block of bytes for a slot.
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * buf <-> Buffer we are filling with the random bytes.
 *
 * slot_p <-> Pointer to slot which will be updated with the random block.
 *
 * max_alloc -> Maximum space per allocation.
 */
static	void	create_block(char *buf, test_slot_t *slot_p,
			     const int max_alloc)
{
  char		*buf_p, *bounds_p;
  unsigned int	left;
  int		stuff;
  
  /* create a random chunk */
  do {
    slot_p->ts_data_size = _dh_os_random() % max_alloc;
  } while (slot_p->ts_data_size == 0);
  slot_p->ts_user_type = _dh_os_random();
  
  buf_p = buf;
  bounds_p = buf + slot_p->ts_data_size;
  
  /* fill the buffer with random stuff */
  while (buf_p < bounds_p) {
    
    /* create a random 4 bytes */
    stuff = _dh_os_random();
    
    left = bounds_p - buf_p;
    if (sizeof(stuff) < left) {
      left = sizeof(stuff);
    }
    memmove(buf_p, &stuff, left);
    buf_p += left;
  }
  
  /* get a hash value for the data */
  slot_p->ts_data_hash = hash(buf, slot_p->ts_data_size, 0);
}

/*
 * static void test_initial
 *
 * DESCRIPTION:
 *
 * Test initial create/open/close functions.
 *
 * RETURNS:
 * 
 * Success - 1
 *
 * Failure - 0
 *
 * ARGUMENTS:
 *
 * None.
 */
static	int	test_initial(void)
{
  int		ret;
  unsigned int	create_type, open_type;
  diskheap_t	*dh_p, *dh2_p;
  
  (void)unlink(TEST_DB_FILE);
  if (verbose_b) {
    (void)printf("unlinked '%s'\n", TEST_DB_FILE);
  }
  
  /* some random number */
  create_type = _dh_os_random();
  
  /* create our heap */
  dh_p = diskheap_create(TEST_DB_FILE, 0 /* flags */, 0 /* block-size */,
			 create_type, 0666 /* open-mode */, &ret);
  if (dh_p == NULL) {
    if (! silent_b) {
      (void)fprintf(error_stream,
		    "%s: could not diskheap_create: %s (%d): %s\n",
		    program, diskheap_strerror(ret), ret, strerror(errno));
    }
    return 0;
  }
  if (verbose_b) {
    (void)printf("created '%s' with heap-type %d\n",
		 TEST_DB_FILE, create_type);
  }
  
  /* close the heap */
  ret = diskheap_close(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: diskheap_close failed: %s (%d)\n",
		    program, diskheap_strerror(ret), ret);
    }
    return 0;
  }
  
  /****************/
  
  /* re-open the heap */
  dh_p = diskheap_open(TEST_DB_FILE, 0 /* flags */, &open_type, &ret);
  if (dh_p == NULL) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: could not diskheap_open: %s (%d): %s\n",
		    program, diskheap_strerror(ret), ret, strerror(errno));
    }
    return 0;
  }
  
  /* the type should be the same in the open as it was in the create */
  if (open_type != create_type) {
    if (! silent_b) {
      (void)fprintf(error_stream,
		    "%s: diskheap_open user-type %d doesn't match create %d\n",
		    program, open_type, create_type);
    }
    (void)diskheap_close(dh_p);
    return 0;
  }
  if (verbose_b) {
    (void)printf("reopened %s, got heap-type %d\n", TEST_DB_FILE, open_type);
  }
  
  /* close the heap */
  ret = diskheap_close(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: diskheap_close failed: %s (%d)\n",
		    program, diskheap_strerror(ret), ret);
    }
    return 0;
  }
  
  /****************/
  
  /* re-open the heap */
  dh_p = diskheap_open(TEST_DB_FILE, 0 /* flags */, &open_type, &ret);
  if (dh_p == NULL) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: could not diskheap_open: %s (%d): %s\n",
		    program, diskheap_strerror(ret), ret, strerror(errno));
    }
    return 0;
  }
  if (verbose_b) {
    (void)printf("reopened %s, got heap-type %d\n", TEST_DB_FILE, open_type);
  }
  
  /* trying to open it again without a close */
  
  /* re-open the heap */
  dh2_p = diskheap_open(TEST_DB_FILE, 0 /* flags */, &open_type, &ret);
  if (dh2_p != NULL) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: opened '%s' twice returned: %s (%d)\n",
		    program, TEST_DB_FILE, diskheap_strerror(ret), ret);
    }
    return 0;
  }
  
  if (ret != DISKHEAP_ERROR_BAD_CLOSE) {
    if (! silent_b) {
      (void)fprintf(error_stream,
		    "%s: opened '%s' without close but got: %s (%d)\n",
		    program, TEST_DB_FILE, diskheap_strerror(ret), ret);
    }
    return 0;
  }
  
  if (verbose_b) {
    (void)printf("re-opened %s without closing, got %s (%d)\n",
		 TEST_DB_FILE, diskheap_strerror(ret), ret);
  }
  (void)diskheap_close(dh_p);
  
  return 1;
}

/*
 * static void test_basic
 *
 * DESCRIPTION:
 *
 * Test store and retrieve functions
 *
 * RETURNS:
 * 
 * Success - 1
 *
 * Failure - 0
 *
 * ARGUMENTS:
 *
 * None.
 */
static	int	test_basic(void)
{
  int		ret;
  unsigned int	create_type, read_type, read_size, test_len;
  unsigned int	block_num, offset, block_num2, offset2;
  diskheap_t	*dh_p;
  char		*test_string, *read_string;
  
  (void)unlink(TEST_DB_FILE);
  if (verbose_b) {
    (void)printf("unlinked '%s'\n", TEST_DB_FILE);
  }
  
  /* create our heap */
  dh_p = diskheap_create(TEST_DB_FILE, 0 /* flags */, 0 /* block-size */,
			 0 /* heap type */, 0666 /* open-mode */, &ret);
  if (dh_p == NULL) {
    if (! silent_b) {
      (void)fprintf(error_stream,
		    "%s: could not diskheap_create: %s (%d): %s\n",
		    program, diskheap_strerror(ret), ret, strerror(errno));
    }
    return 0;
  }
  if (verbose_b) {
    (void)printf("created '%s'\n", TEST_DB_FILE);
  }
  
  test_string = "Hello there folks.  This is a test.  Remain calm.";
  test_len = strlen(test_string);
  create_type = _dh_os_random();
  
  /* store our test string */
  ret = diskheap_store(dh_p, test_string, test_len, create_type, &block_num,
		       &offset);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: diskheap_store failed: %s (%d)\n",
		    program, diskheap_strerror(ret), ret);
    }
    (void)diskheap_close(dh_p);
    return 0;
  }
  if (verbose_b) {
    (void)printf("stored %d bytes, user-type %u, block %u/%u\n",
		 test_len, create_type, block_num, offset);
  }
  
  /* retrieve our test string */
  read_string = diskheap_retrieve(dh_p, block_num, offset,
				  &read_size, &read_type, &ret);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: diskheap_retrieve failed: %s (%d)\n",
		    program, diskheap_strerror(ret), ret);
    }
    (void)diskheap_close(dh_p);
    return 0;
  }
  
  if (read_size != test_len) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: read-size of %u != stored size %u\n",
		    program, read_size, test_len);
    }
    return 0;
  }
  if (read_type != create_type) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: read-type of %u != stored type %u\n",
		    program, read_type, create_type);
    }
    return 0;
  }
  if (memcmp(test_string, read_string, test_len) != 0) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: read-string != stored string '%s'\n",
		    program, test_string);
    }
    return 0;
  }
  free(read_string);
  
  if (verbose_b) {
    (void)printf("read %d bytes, user-type %u, which matched the stored one\n",
		 test_len, read_type);
  }
  
  /* close the heap */
  ret = diskheap_close(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: diskheap_close failed: %s (%d)\n",
		    program, diskheap_strerror(ret), ret);
    }
    return 0;
  }
  
  /****************/
  
  /* re-open the heap */
  dh_p = diskheap_open(TEST_DB_FILE, 0 /* flags */, NULL, &ret);
  if (dh_p == NULL) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: could not diskheap_open: %s (%d): %s\n",
		    program, diskheap_strerror(ret), ret, strerror(errno));
    }
    return 0;
  }
  if (verbose_b) {
    (void)printf("reopened %s\n", TEST_DB_FILE);
  }

  /* retrieve our test string */
  read_string = diskheap_retrieve(dh_p, block_num, offset,
				  &read_size, &read_type, &ret);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: diskheap_retrieve failed: %s (%d)\n",
		    program, diskheap_strerror(ret), ret);
    }
    (void)diskheap_close(dh_p);
    return 0;
  }
  
  if (read_size != test_len) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: read-size of %u != stored size %u\n",
		    program, read_size, test_len);
    }
    return 0;
  }
  if (read_type != create_type) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: read-type of %u != stored type %u\n",
		    program, read_type, create_type);
    }
    return 0;
  }
  if (memcmp(test_string, read_string, test_len) != 0) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: read-string != stored string '%s'\n",
		    program, test_string);
    }
    return 0;
  }
  free(read_string);
  
  if (verbose_b) {
    (void)printf("read %d bytes, user-type %u, which matched the stored one\n",
		 test_len, read_type);
  }
  
  /****************/
  
  /* now delete the entry */
  ret = diskheap_delete(dh_p, block_num, offset, &read_size, &read_type);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (! silent_b) {
      (void)fprintf(error_stream,
		    "%s: diskheap_delete of block %u/%u: %s (%d)\n",
		    program, block_num, offset, diskheap_strerror(ret), ret);
    }
    return 0;
  }
  if (read_size != test_len) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: read-size of %u != stored size %u\n",
		    program, read_size, test_len);
    }
    return 0;
  }
  if (read_type != create_type) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: read-type of %u != stored type %u\n",
		    program, read_type, create_type);
    }
    return 0;
  }
  
  if (verbose_b) {
    (void)printf("deleted entry of %d bytes, block %u/%u\n",
		 test_len, block_num, offset);
  }
  
  /* close the heap */
  ret = diskheap_close(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: diskheap_close failed: %s (%d)\n",
		    program, diskheap_strerror(ret), ret);
    }
    return 0;
  }
  
  /****************/
  /* test the free list */ 
  
  /* re-open the heap */
  dh_p = diskheap_open(TEST_DB_FILE, 0 /* flags */, NULL, &ret);
  if (dh_p == NULL) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: could not diskheap_open: %s (%d): %s\n",
		    program, diskheap_strerror(ret), ret, strerror(errno));
    }
    return 0;
  }
  if (verbose_b) {
    (void)printf("reopened %s\n", TEST_DB_FILE);
  }
  
  /* store our test string */
  ret = diskheap_store(dh_p, test_string, test_len, create_type, &block_num,
		       &offset);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: diskheap_store failed: %s (%d)\n",
		    program, diskheap_strerror(ret), ret);
    }
    (void)diskheap_close(dh_p);
    return 0;
  }
  if (verbose_b) {
    (void)printf("stored %d bytes, user-type %u, block %u/%u\n",
		 test_len, create_type, block_num, offset);
  }
  
  /* now delete the entry */
  ret = diskheap_delete(dh_p, block_num, offset, &read_size, &read_type);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (! silent_b) {
      (void)fprintf(error_stream,
		    "%s: diskheap_delete of block %u/%u: %s (%d)\n",
		    program, block_num, offset, diskheap_strerror(ret), ret);
    }
    return 0;
  }
  if (read_size != test_len) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: read-size of %u != stored size %u\n",
		    program, read_size, test_len);
    }
    return 0;
  }
  if (read_type != create_type) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: read-type of %u != stored type %u\n",
		    program, read_type, create_type);
    }
    return 0;
  }
  
  if (verbose_b) {
    (void)printf("deleted entry of %d bytes, block %u/%u\n",
		 test_len, block_num, offset);
  }
  
  /* now store it again */
  ret = diskheap_store(dh_p, test_string, test_len, create_type, &block_num2,
		       &offset2);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: diskheap_store failed: %s (%d)\n",
		    program, diskheap_strerror(ret), ret);
    }
    (void)diskheap_close(dh_p);
    return 0;
  }
  if (verbose_b) {
    (void)printf("stored %d bytes, user-type %u, block %u/%u\n",
		 test_len, create_type, block_num2, offset2);
  }
  if (block_num != block_num2 || offset != offset2) {
    if (! silent_b) {
      (void)fprintf(error_stream,
		    "%s: store in %u/%u, delete, store did not put at %u/%u\n",
		    program, block_num, offset, block_num2, offset2);
    }
    (void)diskheap_close(dh_p);
    return 0;
  }
  
  /* close the heap */
  ret = diskheap_close(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: diskheap_close failed: %s (%d)\n",
		    program, diskheap_strerror(ret), ret);
    }
    return 0;
  }
  
  return 1;
}

/*
 * static void test_random
 *
 * DESCRIPTION:
 *
 * Test a large number of random store/retrieves.
 *
 * RETURNS:
 * 
 * Success - 1
 *
 * Failure - 0
 *
 * ARGUMENTS:
 *
 * block_size -> Block-size of created heaps.
 *
 * iter_n -> Number of iterations to run.
 *
 * max_alloc -> Maximum space per allocation.
 */
static	int	test_random(const int block_size, const int iter_n,
			    const int max_alloc)
{
  diskheap_t	*dh_p;
  test_slot_t	*slots, *slot_p;
  char		*buf, *buf_p;
  unsigned int	read_size, read_type, read_hash;
  int		iter_c, ret, which, slot_n, final = 1;
  
  /* allocate our buffer */
  buf = malloc(max_alloc);
  if (buf == NULL) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: could not allocate buffer space: %s\n",
		    program, strerror(errno));
    }
    return 0;
  }
  
  /*
   * allocate our slot array with 1/10th of the number of items that
   * we will be iterating across
   */
  slot_n = iter_n / 10;
  slots = calloc(slot_n, sizeof(*slots));
  if (slots == NULL) {
    free(buf);
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: could not allocate slot array: %s\n",
		    program, strerror(errno));
    }
    return 0;
  }
  
  (void)unlink(TEST_DB_FILE);
  if (verbose_b) {
    (void)printf("unlinked '%s'\n", TEST_DB_FILE);
  }
  
  /* create our heap */
  dh_p = diskheap_create(TEST_DB_FILE, 0 /* flags */, block_size,
			 0 /* heap type */, 0666 /* open-mode */, &ret);
  if (dh_p == NULL) {
    if (! silent_b) {
      (void)fprintf(error_stream,
		    "%s: could not diskheap_create: %s (%d): %s\n",
		    program, diskheap_strerror(ret), ret, strerror(errno));
    }
    free(buf);
    free(slots);
    return 0;
  }
  if (verbose_b) {
    (void)printf("created '%s'\n", TEST_DB_FILE);
  }
  
  for (iter_c = 0; iter_c < iter_n; iter_c++) {
    
    /* 1/100th of the time we reopen the file */
    if (_dh_os_random() % 100 == 50) {
      
      /* this is where we re-open the file */
      ret = diskheap_close(dh_p);
      if (ret != DISKHEAP_ERROR_NONE) {
	if (! silent_b) {
	  (void)fprintf(error_stream, "%s: diskheap_close failed: %s (%d)\n",
			program, diskheap_strerror(ret), ret);
	}
	free(buf);
	free(slots);
	return 0;
      }
      
      /* now re-open it */
      dh_p = diskheap_open(TEST_DB_FILE, 0 /* flags */, NULL /* heap-type */,
			   &ret);
      if (dh_p == NULL) {
	if (! silent_b) {
	  (void)fprintf(error_stream,
			"%s: could not diskheap_open '%s': %s (%d): %s\n",
			program, TEST_DB_FILE, diskheap_strerror(ret), ret,
			strerror(errno));
	}
	free(buf);
	free(slots);
	return 0;
      }
      
      if (verbose_b) {
	(void)printf("i%d: reopened '%s'\n", iter_c, TEST_DB_FILE);
      }
    }
    
    /* what should we do */
    which = _dh_os_random() % slot_n;
    slot_p = slots + which;
    
    /* is the slot free? */
    if (slot_p->ts_block_num == 0) {
      
      /*
       * half of the time allocate a small chunk to test the sub-block
       * sizes
       */
      if (_dh_os_random() % 10 > 5) {
	if (block_size == 0) {
	  create_block(buf, slot_p, 4096);
	}
	else {
	  create_block(buf, slot_p, block_size);
	}
      }
      else {
	create_block(buf, slot_p, max_alloc);
      }
      
      /* store data to the db */
      ret = diskheap_store(dh_p, buf, slot_p->ts_data_size,
			   slot_p->ts_user_type, &slot_p->ts_block_num,
			   &slot_p->ts_offset);
      if (ret != DISKHEAP_ERROR_NONE) {
	if (! silent_b) {
	  (void)fprintf(error_stream, "%s: diskheap_store failed: %s (%d)\n",
			program, diskheap_strerror(ret), ret);
	}
	(void)diskheap_close(dh_p);
	free(buf);
	free(slots);
	return 0;
      }
      if (verbose_b) {
	(void)printf("i%d: s%d: stored %d bytes at block %u/%u\n",
		     iter_c, slot_p - slots, slot_p->ts_data_size,
		     slot_p->ts_block_num, slot_p->ts_offset);
      }
      continue;
    }
    
    /*
     * slot already is full
     */
    
    /* see whether we should read it, delete it, or update it */
    which = _dh_os_random() % 5;
    
    /* delete it 1/5 of the time */
    if (which == 0) {
      
      /* delete the data from the db */
      ret = diskheap_delete(dh_p, slot_p->ts_block_num, slot_p->ts_offset,
			    &read_size, &read_type);
      if (ret == DISKHEAP_ERROR_NONE) {
	if (read_size != slot_p->ts_data_size) {
	  if (! silent_b) {
	    (void)fprintf(error_stream,
			  "%s: delete size %u doesn't match stored %u\n",
			  program, read_size, slot_p->ts_data_size);
	  }
	  final = 0;
	}
	if (read_type != slot_p->ts_user_type) {
	  if (! silent_b) {
	    (void)fprintf(error_stream,
			  "%s: delete type %u doesn't match stored %u\n",
			  program, read_type, slot_p->ts_user_type);
	  }
	  final = 0;
	}
	if (verbose_b) {
	  (void)printf("i%d: s%d: deleted %d bytes from block %u/%u\n",
		       iter_c, slot_p - slots, read_size, slot_p->ts_block_num,
		       slot_p->ts_offset);
	}
	slot_p->ts_block_num = 0;
	slot_p->ts_offset = 0;
      }
      else {
	if (! silent_b) {
	  (void)fprintf(error_stream,
			"%s: diskheap_delete of %u/%u failed: %s (%d)\n",
			program, slot_p->ts_block_num, slot_p->ts_offset,
			diskheap_strerror(ret), ret);
	}
	final = 0;
      }
      continue;
    }
    
    /* update the block 2/5 of the time */
    if (which == 1 || which == 2) {
      int		safer_b;
      test_slot_t	new_slot;
      
      /* create a new slot, half of the time a sub-block */
      if (_dh_os_random() % 10 > 5) {
	if (block_size == 0) {
	  create_block(buf, &new_slot, 4096);
	}
	else {
	  create_block(buf, &new_slot, block_size);
	}
      }
      else {
	create_block(buf, &new_slot, max_alloc);
      }
      
      /* either do it safer or not */
      safer_b = _dh_os_random() % 2;
      
      /* update the data in the db */
      ret = diskheap_update(dh_p, buf, new_slot.ts_data_size,
			    new_slot.ts_user_type, slot_p->ts_block_num,
			    slot_p->ts_offset, safer_b,
			    &new_slot.ts_block_num, &new_slot.ts_offset,
			    &read_size, &read_type);
      if (ret != DISKHEAP_ERROR_NONE) {
	if (! silent_b) {
	  (void)fprintf(error_stream, "%s: diskheap_update failed: %s (%d)\n",
			program, diskheap_strerror(ret), ret);
	}
	(void)diskheap_close(dh_p);
	free(buf);
	free(slots);
	return 0;
      }
      if (read_size != slot_p->ts_data_size) {
	if (! silent_b) {
	  (void)fprintf(error_stream,
			"%s: update size %u doesn't match stored %u\n",
			program, read_size, slot_p->ts_data_size);
	}
	final = 0;
      }
      if (read_type != slot_p->ts_user_type) {
	if (! silent_b) {
	  (void)fprintf(error_stream,
			"%s: update type %u doesn't match stored %u\n",
			program, read_type, slot_p->ts_user_type);
	}
	final = 0;
      }
      if (verbose_b) {
	(void)printf("i%d: s%d: updated %s: %d to %d bytes, loc %u/%u to "
		     "%u/%u\n",
		     iter_c, slot_p - slots, (safer_b ? "safe" : "not-safe"),
		     read_size, new_slot.ts_data_size,
		     slot_p->ts_block_num, slot_p->ts_offset,
		     new_slot.ts_block_num, new_slot.ts_offset);
      }
      *slot_p = new_slot;
      continue;
    }
    
    /* retrieve data from the db 2/5 of the time */
    
    /* decide between retrieve and retrive-to-buf */
    if (_dh_os_random() % 2 == 0) {
      /* retrive to buffer */
      buf_p = diskheap_retrieve(dh_p, slot_p->ts_block_num, slot_p->ts_offset,
				&read_size, &read_type, &ret);
      if (buf_p == NULL) {
	if (! silent_b) {
	  (void)fprintf(error_stream,
			"%s: diskheap_retrieve_to_buf of %u/%u failed: "
			"%s (%d)\n",
			program, slot_p->ts_block_num, slot_p->ts_offset,
			diskheap_strerror(ret), ret);
	}
	final = 0;
	continue;
      }
    }
    else {
      /* retrive to buffer */
      ret = diskheap_retrieve_to_buf(dh_p, slot_p->ts_block_num,
				     slot_p->ts_offset, buf, max_alloc,
				     &read_size, &read_type);
      if (ret != DISKHEAP_ERROR_NONE) {
	if (! silent_b) {
	  (void)fprintf(error_stream,
			"%s: diskheap_retrieve_to_buf of %u/%u failed: "
			"%s (%d)\n",
			program, slot_p->ts_block_num, slot_p->ts_offset,
			diskheap_strerror(ret), ret);
	}
	final = 0;
	continue;
      }
      buf_p = NULL;
    }
    if (read_size != slot_p->ts_data_size) {
      if (! silent_b) {
	(void)fprintf(error_stream,
		      "%s: read size %u doesn't match stored %u\n",
		      program, read_size, slot_p->ts_data_size);
      }
      final = 0;
      free(buf_p);
      continue;
    }
    if (read_type != slot_p->ts_user_type) {
      if (! silent_b) {
	(void)fprintf(error_stream,
		      "%s: read type %u doesn't match stored %u\n",
		      program, read_type, slot_p->ts_user_type);
      }
      final = 0;
      free(buf_p);
      continue;
    }
    if (buf_p == NULL) {
      read_hash = hash(buf, read_size, 0);
    }
    else {
      read_hash = hash(buf_p, read_size, 0);
      free(buf_p);
    }
    if (read_hash != slot_p->ts_data_hash) {
      if (! silent_b) {
	(void)fprintf(error_stream,
		      "%s: read hash %u doesn't match stored %u\n",
		      program, read_hash, slot_p->ts_data_hash);
      }
      final = 0;
      continue;
    }
    if (verbose_b) {
      (void)printf("i%d: s%d: read %d bytes from block %u/%u\n",
		   iter_c, slot_p - slots, slot_p->ts_data_size,
		   slot_p->ts_block_num, slot_p->ts_offset);
    }
  }
  
  free(buf);
  free(slots);
  
  /* close the heap */
  ret = diskheap_close(dh_p);
  if (ret != DISKHEAP_ERROR_NONE) {
    if (! silent_b) {
      (void)fprintf(error_stream, "%s: diskheap_close failed: %s (%d)\n",
		    program, diskheap_strerror(ret), ret);
    }
    final = 0;
  }
  
  return final;
}

/*
 * static void usage
 *
 * DESCRIPTION:
 *
 * Print usage message.
 *
 * RETURNS:
 * 
 * None.
 *
 * ARGUMENTS:
 *
 * None.
 */
static	void	usage(void)
{
  (void)printf("Usage: %s [-b size] [-i number] [-m max-alloc] [-s] [-S seed] "
	       "[-v]\n",
	       program);
  (void)printf("   -b size    - block-size in bytes for the heap\n");
  (void)printf("   -i number  - number of iterations to run (default %d)\n",
	       DEFAULT_ITERATIONS);
  (void)printf("   -m number  - maximum per-allocation size (default %d)\n",
	       MAX_ALLOCATION);
  (void)printf("   -s         - silent operation\n");
  (void)printf("   -S seed    - specify a random seed value to use\n");
  (void)printf("   -v         - verbose messages\n");
}

int	main(int argc, char **argv)
{
  int	seed, iter_n = DEFAULT_ITERATIONS, max_alloc = MAX_ALLOCATION;
  int	block_size = 0;
  
  program = *argv;
  
  seed = time(NULL);
#if HAVE_GETPID
  seed ^= getpid();
#endif
  
  for (argc--, argv++; argc > 0 && **argv == '-'; argc--, argv++) {
    if (*(*argv + 1) == '\0') {
      break;
    }
    if (*(*argv + 1) == 'b') {
      argc--, argv++;
      if (argc == 0) {
	usage();
      }
      block_size = atoi(*argv);
      continue;
    }
    if (*(*argv + 1) == 'i') {
      argc--, argv++;
      if (argc == 0) {
	usage();
      }
      iter_n = atoi(*argv);
      continue;
    }
    if (*(*argv + 1) == 'm') {
      argc--, argv++;
      if (argc == 0) {
	usage();
      }
      max_alloc = atoi(*argv);
      continue;
    }
    if (*(*argv + 1) == 's') {
      silent_b = 1;
      continue;
    }
    if (*(*argv + 1) == 'S') {
      argc--, argv++;
      if (argc == 0) {
	usage();
      }
      seed = atoi(*argv);
      continue;
    }
    if (*(*argv + 1) == 'v') {
      verbose_b = 1;
      continue;
    }
    
    usage();
    return 1;
  }
  
  if (argc > 0) {
    usage();
    return 1;
  }
  
  /* verbose wins */
  if (silent_b && verbose_b) {
    silent_b = 0;
  }
  
  if (verbose_b) {
    error_stream = stdout;
  }
  else {
    error_stream = stderr;
    if (! silent_b) {
      (void)printf("Running tests.  Specify the -v (--verbose) flag for more "
		   "details\n");
    }
  }
  
  diskheap_seed_random(seed);
  if (! silent_b) {
    (void)printf("Random seed is %d\n", seed);
  }
  
  /* initial create/open/close tests */
  if ((! test_initial())
      || (! test_basic())
      || (! test_random(block_size, iter_n, max_alloc))) {
    if (silent_b) {
      (void)fprintf(error_stream, "Testing failed.  Random seed is %u.\n",
		    seed);
    }
    return 1;
  }
  
  return 0;
}
